/*
** Module   : SaxReader
** Abstract : SAX reader specific interface.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130208 Created initial version.
** 002 CA  20130221 Added ADM-DATA and UNIQUE-ID support. 
** 003 EVL 20130320 Adding SAX reader related literal constants.
** 004 EVK 20130903 Moved SuppressNamespaceProcessing methods to the new interface XmlSchema.
** 005 CA  20130927 Added LegacyResource annotation, to determine resource type.
** 006 EVL 20130911 Adding legacy attributes and methods annotation.
** 007 EVL 20131009 Leaving only SAX specific methods for schema path and schema locations. Other
**                  are inherited from XmlSchema classs.
** 008 EVK 20131001 Added enum with 4gl reader status.
*/

package com.goldencode.p2j.xml;

import com.goldencode.p2j.util.*;

/**
 * Interface describing SAX-Reader specific features of the SAX objects.
 */
@LegacyResource(resource = LegacyResource.SAX_READER)
public interface SaxReader
extends SaxEntity,
        UniqueID,
        ADMData,
        XmlSchema
{
   /**
    * List of possible states of SAX-READER object, returned by the PARSE-STATUS attribute:
    * <ul>
    * <li>{@link #SAX_UNINITIALIZED} -The SAX-UNINITIALIZED constant which in 4GL is the integer
    * value 1.</li>
    * <li>{@link #SAX_RUNNING} - The SAX-RUNNING constant which in 4GL is the integer 2.</li>
    * <li>{@link #SAX_COMPLETE} - The SAX-COMPLETE constant which in 4GL is the integer 3.</li>
    * <li>{@link #SAX_PARSER_ERROR} - The SAX-PARSER-ERROR constant which in 4GL is the integer 4.
    * </li>
    * </ul>
    */
   enum ReadStatus
   {
      SAX_UNINITIALIZED(1, "SAX-UNINITIALIZED"),
      SAX_RUNNING      (2, "SAX-RUNNING"),
      SAX_COMPLETE     (3, "SAX-COMPLETE"),
      SAX_PARSER_ERROR (4, "SAX-PARSER-ERROR");

      /** 4gl constant value */
      private final int    status;
      /** 4gl constant name */
      private final String name;

      ReadStatus(int status, String name)
      {
         this.status = status;
         this.name = name;
      }

      /**
       * Returns the 4gl name of this enum constant.
       *
       * @return See above.
       */
      public String getName()
      {
         return name;
      }

      /**
       * Returns the 4gl value of this enum constant.
       *
       * @return See above.
       */
      public int getStatus()
      {
         return status;
      }

      /**
       * Returns the name of this enum constant, as contained in the
       * declaration.  This method may be overridden, though it typically
       * isn't necessary or desirable.  An enum type should override this
       * method when a more "programmer-friendly" string form exists.
       *
       * @return the name of this enum constant
       */
      @Override
      public String toString()
      {
         return name;
      }
   }
   // Attributes

   /**
    * Gets the value of the handler attribute for the given SAX Reader object.
    *
    * @return   The handle variable representing the handler attribute.
    */
   @LegacyAttribute(name = "HANDLER")
   public handle getHandler();

   /**
    * Sets the value of the handler attribute for the given SAX Reader object.
    *
    * @param    hCallback
    *           The handle of the new callback handler to be set for the handler attribute.
    */
   @LegacyAttribute(name = "HANDLER", setter = true)
   public void setHandler(handle hCallback);

   /**
    * Gets the value of the current 1-based column in XML source.
    *
    * @return   The integer value of the current column.
    */
   @LegacyAttribute(name = "LOCATOR-COLUMN-NUMBER")
   public integer getLocatorColumnNumber();

   /**
    * Gets the value of the current 1-based line number in XML source.
    *
    * @return   The integer value of the current line.
    */
   @LegacyAttribute(name = "LOCATOR-LINE-NUMBER")
   public integer getLocatorLineNumber();

   /**
    * Gets the current value of the parse status. The default value is
    * {@link ReadStatus#SAX_UNINITIALIZED}. The other returned values are
    * {@link ReadStatus#SAX_COMPLETE}, {@link ReadStatus#SAX_PARSER_ERROR} or
    * {@link ReadStatus#SAX_RUNNING}.
    *
    * @return   The integer value representing the current status of the parse.
    */
   @LegacyAttribute(name = "PARSE-STATUS")
   public integer getParseStatus();

   /**
    * Gets the current value of the public identifier of the current XML source.
    *
    * @return   The character value representing the current locator-public-id attribute.
    */
   @LegacyAttribute(name = "LOCATOR-PUBLIC-ID")
   public character getLocatorPublicId();

   /**
    * Gets the current value of the system identifier of the current XML source.
    * 
    * @return   The character value representing the current locator-system-id attribute.
    */
   @LegacyAttribute(name = "LOCATOR-SYSTEM-ID")
   public character getLocatorSystemId();

   /**
    * Sets the current value of the namespace location pair mapping.
    * 
    * @param    location
    *           The new value the current schema-location attribute.
    */
   @LegacyAttribute(name = "SCHEMA-LOCATION", setter = true)
   public void setSchemaLocation(String location);
   
   /**
    * Sets the delimiter separated list of directory paths used to locate XML DTD associated with
    * a particular XML document.
    * 
    * @param    path
    *           The new value the current schema-path attribute.
    */
   @LegacyAttribute(name = "SCHEMA-PATH", setter = true)
   public void setSchemaPath(String path);
   
   /**
    * Indicates if the parser validates the XML document against DTD or not.
    *
    * @return   The logical value representing the current validation-enabled attribute.
    */
   @LegacyAttribute(name = "VALIDATION-ENABLED")
   public logical getValidationEnabled();
   
   /**
    * Redefines whether the parser will validate the XML document against DTD or not.
    * 
    * @param    enable
    *           The new value the current validation-enabled attribute.
    */
   @LegacyAttribute(name = "VALIDATION-ENABLED", setter = true)
   public void setValidationEnabled(logical enable);
   
   /**
    * Redefines whether the parser will validate the XML document against DTD or not.
    * 
    * @param    enable
    *           The new value the current validation-enabled attribute.
    */
   @LegacyAttribute(name = "VALIDATION-ENABLED", setter = true)
   public void setValidationEnabled(boolean enable);
   
   // Methods
   /**
    * Executing single call parse of an XML document associated with a SAX-Reader object.
    */
   @LegacyMethod(name = "SAX-PARSE")
   public void saxParse();

   /**
    * Initializes and begins progressive-scan parse of an XML document associated with a
    * SAX-Reader object.
    */
   @LegacyMethod(name = "SAX-PARSE-FIRST")
   public void saxParseFirst();

   /**
    * Continues progressive-scan parse of an XML document associated with a SAX-Reader object.
    */
   @LegacyMethod(name = "SAX-PARSE-NEXT")
   public void saxParseNext();

   /**
    * Specifies the input source of the XML document to be parsed by a SAX-Reader object.
    * 
    * @param    mode
    *           The type of the object to process. Can be 'FILE', 'HANDLE', 'MEMPTR' or
    *           'LONGCHAR'.
    * @param    source
    *           The appropriate object to process.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @LegacyMethod(name = "SET-INPUT-SOURCE")
   public logical setInputSource(String mode, Object source);

   /**
    * Specifies the input source of the XML document to be parsed by a SAX-Reader object.
    * 
    * @param    mode
    *           The type of the object to process. Can be 'FILE', 'HANDLE', 'MEMPTR' or
    *           'LONGCHAR'.
    * @param    source
    *           The appropriate object to process.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @LegacyMethod(name = "SET-INPUT-SOURCE")
   public logical setInputSource(character mode, Object source);
   
   /**
    * Causes parser to stop parsing the XML document.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @LegacyMethod(name = "STOP-PARSING")
   public logical stopParsing();
}
