/*
** Module   : XEntityImpl
** Abstract : DOM and SAX common implementation.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130204 Created initial version. Major rework to use w3c backend.
** 002 EVL 20130214 Adding the base inheritance from XCommonImpl class to place the XML related
**                  common features.  
** 003 CA  20130221 Added UNIQUE-ID and NAMESPACE-URI support. 
** 004 EVL 20130222 Adding SUBTYPE attribute and INSERT-BEFORE method support.
** 005 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 006 OM  20130320 Fixed getChild() method, added Override annotations, cosmetics & typo fixes.
** 007 EVK 20130527 Adding support CHILD-NUM, NAMESPACE-PREFIX attributes and REPLACE-CHILD
**                  method.
**                  Add implementation for SUBTYPE, NUM-CHILDREN, CHILD-NUM, UNIQUE-ID,
**                  NAMESPACE-URI, NAME, NAMESPACE-PREFIX attributes and IMPORT-NODE,
**                  APPEND-CHILD, INSERT-BEFORE,REMOVE-CHILD, REPLACE-CHILD, GET-CHILD,
**                  REMOVE-ATTRIBUTE methods.
**                  Remove map Element->handle, because we don't need to store handle
**                  reference for children.
**                  xnode variable is now Node instead of Element.
** 008 CA  20130927 Removed unknown() API, as by default the resource is known. 
** 009 EVK 20131001 Moved up to hierarchy next methods:
**                  invalidArgumentError, invalidArgumentAssignError.
*/

package com.goldencode.p2j.xml;

import org.w3c.dom.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of the common features for DOM.
 */
public abstract class XEntityImpl
extends XCommonImpl
implements XEntity
{
   /** Element associates with current XEntity class. */
   private Node xnode = null;

   /**
    * Default constructor.
    */
   protected XEntityImpl()
   {
   }

   /**
    * Special constructor accepting a X-Document or X-Node-Reference object.
    *
    * @param    node
    *           The DOM XML node object to be associated with the given
    *           handle object.
    */
   protected XEntityImpl(Node node)
   {
      xnode = node;
   }

   /**
    * Utility method to set up node value into handle. If handle is not XNodeRef or XDocument
    * type then do nothing.
    *
    * @param    handle
    *           The handle class instance to set new node value.
    * @param    node
    *           The node value to set.
    */
   static void setNodeToHandle(handle handle, Node node)
   {
      if (isXEntityHandle(handle))
      {
         ((XEntityImpl)handle.getResource()).setNode(node);
      }
   }

   /**
    * Check the type of resource in handle. Return <code>true</code> if and only if handle
    * contains resource of type XNodeRef.
    *
    * @param    handle
    *           The handle class instance to check type.
    *
    * @return   See above.
    */
   static boolean isXNodeHandle(handle handle)
   {
      if (handle == null || !handle._isValid() || handle.isUnknown())
      {
         return false;
      }
      WrappedResource resource = handle.getResource();
      return resource instanceof XNodeRefImpl;
   }

   /**
    * Check the type of resource in handle. Return <code>true</code> if and only if handle
    * contains resource of type XNodeRef or XDocument.
    *
    * @param    handle
    *           The handle class instance to check type.
    *
    * @return   See above.
    */
   static boolean isXEntityHandle(handle handle)
   {
      if (handle == null || !handle._isValid() || handle.isUnknown())
      {
         return false;
      }
      WrappedResource resource = handle.getResource();
      return resource instanceof XEntityImpl;
   }

   /**
    * Check the type of resource in handle. Return <code>true</code> if and only if handle
    * contains resource of type XNodeRef.
    *
    * @param    handle
    *           The handle class instance to check type.
    *
    * @return   See above.
    */
   static boolean isXNodeHandleInitialized(handle handle)
   {
      if (!isXNodeHandle(handle))
      {
         return false;
      }
      WrappedResource resource = handle.getResource();
      return ((XEntityImpl) resource).isInitialized();
   }

   /**
    * Getting the Element backend object for this XEntity.
    * 
    * @return   The Element associated with object.
    */
   public Node getNode()
   {
      return xnode;
   }
   
   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   @Override
   public boolean valid()
   {
      return true;
   }

   /**
    * Gets the the subtype of the object.
    * 
    * @return   The character value of the subtype attribute.
    */
   @Override
   public character getSubType()
   {
      if (!isInitialized())
      {
         invalidNodeError("SUBTYPE");
         return new character();
      }
      Node node = getNode();
      return new character(XmlFactory.convertNodeToAblNode(node.getNodeType()));
   }
   
   /**
    * Sets the the subtype of the object.
    * 
    * @param    value
    *           The new value of the subtype object attribute.
    */
   @Override
   public void setSubType(character value)
   {
      readOnlyError("SUBTYPE");
   }

   /**
    * Sets the the subtype of the object.
    * 
    * @param    value
    *           The new value of the subtype object attribute.
    */
   @Override
   public void setSubType(String value)
   {
      setSubType(new character(value));
   }

   /**
    * Inserts a node as a child of this document before another node or last if other node is
    * unknown.
    * 
    * @param    handle1
    *           The handle representing the child to insert.
    * @param    handle2
    *           The handle representing the XML node to be inserted before.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical insertBefore(handle handle1, handle handle2)
   {
      if (!isInitialized())
      {
         invalidNodeError("INSERT-BEFORE");
         return new logical(false);
      }

      if (!isXNodeHandle(handle1) || !handle2.isUnknown() && !isXNodeHandle(handle2))
      {
         invalidArgumentError("INSERT-BEFORE");
         return new logical(false);
      }

      if (handle1.isUnknown() || !handle2.isUnknown() && !isXNodeHandleInitialized(handle2))
      {
         invalidNodeError("INSERT-BEFORE");
         return new logical(false);
      }

      if (handle2.isUnknown())
      {
         return appendChild(handle1);
      }
      Node node1 = ((XEntityImpl) handle1.get()).getNode();
      Node node2 = ((XEntityImpl) handle2.get()).getNode();
      try
      {
         getNode().insertBefore(node1, node2);
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.WRONG_DOCUMENT_ERR)
         {
            commonDomError("INSERT-BEFORE",
                           "A node was used in a different document than the one the " +
                           "created it");
         }
         else if (e.code == DOMException.HIERARCHY_REQUEST_ERR)
         {
            commonDomError("INSERT-BEFORE",
                           "An attempt was made to reference a node in a context where it " +
                           "does not exist");
         }
         else if (e.code == DOMException.NOT_FOUND_ERR)
         {
            commonDomError("INSERT-BEFORE", "A node was inserted where it doesn't belong");
         }
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Gets the the number of children objects for the current one.
    * 
    * @return   The integer number of the children.
    */
   @Override
   public integer getNumChildren()
   {
      if (!isInitialized())
      {
         invalidNodeError("NUM-CHILDREN");
         return new integer();
      }

      return new integer(getNode().getChildNodes().getLength());
   }

   /**
    * Removing the specified child from the node.
    *
    * @param    newChildHandle
    *           The handle class instance that represents the node to insert in the tree.
    * @param    oldChildHandle
    *           The handle class instance that represents the node to remove from the tree.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical replaceChild(handle newChildHandle, handle oldChildHandle)
   {
      if (!isInitialized())
      {
         invalidNodeError("REPLACE-CHILD");
         return new logical(false);
      }
      if (!newChildHandle._isValid())
      {
         invalidArgumentError("REPLACE-CHILD");
         return new logical(false);
      }
      if (newChildHandle.isUnknown())
      {
         invalidNodeError("REPLACE-CHILD");
         return new logical(false);
      }
      if (!oldChildHandle._isValid())
      {
         invalidArgumentError("REPLACE-CHILD");
         return new logical(false);
      }
      if (oldChildHandle.isUnknown())
      {
         invalidNodeError("REPLACE-CHILD");
         return new logical(false);
      }


      Node newChild = ((XEntityImpl) newChildHandle.get()).getNode();
      Node oldChild = ((XEntityImpl) oldChildHandle.get()).getNode();
      try
      {
         getNode().replaceChild(newChild, oldChild);
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.WRONG_DOCUMENT_ERR)
         {
            commonDomError("REPLACE-CHILD",
                           "A node was used in a different document than the one " +
                           "the created it");
         }
         else if (e.code == DOMException.HIERARCHY_REQUEST_ERR)
         {
            commonDomError("REPLACE-CHILD", "A node was inserted where doesn't belong");
         }
         else if (e.code == DOMException.NOT_FOUND_ERR)
         {
            commonDomError("REPLACE-CHILD",
                           "An attempt was made to reference a node in a context where " +
                           "it does not exist");
         }

         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Appending the new child to the current object.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be appended.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical appendChild(handle hChild)
   {
      if (!isInitialized() || !valid())
      {
         return new logical(false);
      }

      if (!isXNodeHandle(hChild))
      {
         invalidArgumentError("APPEND-CHILD");
         return new logical(false);
      }

      if (!isXNodeHandleInitialized(hChild))
      {
         invalidNodeError("APPEND-CHILD");
         return new logical(false);
      }

      Node node = getNode();
      Node childNode = ((XEntityImpl) hChild.get()).getNode();
      try
      {
          node.appendChild(childNode);
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.WRONG_DOCUMENT_ERR)
         {
            commonDomError("APPEND-CHILD",
                           "A node was used in a different document than the one the " +
                           "created it");
         }
         else if (e.code == DOMException.HIERARCHY_REQUEST_ERR)
         {
            commonDomError("APPEND-CHILD", "A node was inserted where it doesn't belong");
         }
         else if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
         {
            commonDomError("APPEND-CHILD",
                           "An attempt was made to modify an object where modifications " +
                           "are not allowed");
         }
         return new logical(false);
      }
      return new logical(true);
   }

   /**
    * Getting the handle object of the child node with specified index.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to put
    *           the results.
    * @param    index
    *           1-based index of the child node to be retrieved.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical getChild(handle hChild, int index)
   {
      return getChild(hChild, new integer(index));
   }
   
   /**
    * Getting the handle object of the child node with specified index.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to put the results.
    * @param    index
    *           1-based index of the child node to be retrieved.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical getChild(handle hChild, integer index)
   {
      if (!isInitialized() || !isXNodeHandle(hChild))
      {
         invalidNodeError("GET-CHILD");
         return new logical(false);
      }

      if (!hChild._isValid() || index.isUnknown())
      {
         invalidArgumentError("GET-CHILD");
         return new logical(false);
      }

      int ndx = index.intValue();

      NodeList childNodes = getNode().getChildNodes();
      
      if (ndx < 1 || ndx > childNodes.getLength())
      {
         commonDomError("GET-CHILD", "Invalid child index");
         return new logical(false);
      }
      
      Node node = childNodes.item(ndx - 1);

      setNodeToHandle(hChild, node);

      return new logical(true);
   }

   /**
    * Removing the specified child from the node.
    * 
    * @param    hChild
    *           The valid handle class instance of the child to be removed.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical removeChild(handle hChild)
   {
      if (!isInitialized())
      {
         invalidNodeError("REMOVE-CHILD");
         return new logical(false);
      }

      if (!hChild._isValid())
      {
         invalidArgumentError("REMOVE-CHILD");
         return new logical(false);
      }

      Node elChild = ((XEntityImpl) hChild.get()).getNode();
      try
      {
         getNode().removeChild(elChild);
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.NOT_FOUND_ERR)
         {
            commonDomError("REMOVE-CHILD",
                           "An attempt was made to reference a node in a context where it " +
                           "does not exist");
         }
         else if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
         {
            commonDomError("REMOVE-CHILD",
                           "An attempt was made to modify an object where modifications " +
                           "are not allowed");
         }
         return new logical(false);
      }
      return new logical(true);
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself. For X-noderef it's only unique for current xml Document.
    * 
    * @return   The integer unique number of node inside document.
    */
   @Override
   public integer getUniqueID()
   {
      return new integer(getNode().hashCode());
   }
   
   /**
    * Returns the value of the the NAMESPACE-URI attribute.
    *
    * @return    See above.
    */
   @Override
   public character getNamespaceURI()
   {
      if (!isInitialized())
      {
         invalidNodeError("NAMESPACE-URI");
         return new character();
      }

      String namespaceUri = getNode().getNamespaceURI();
      if (namespaceUri != null)
      {
         return new character(namespaceUri);
      }
      else
      {
         return new character("");
      }
   }

   /**
    * Setter for NAMESPACE-URI attribute. Calls {@link #readOnlyError}, as this attribute is r/o
    * for this resource.
    *
    * @param    uri
    *           Not used.
    */
   @Override
   public void setNamespaceURI(String uri)
   {
      readOnlyError("NAMESPACE-URI");
   }

   /**
    * Setter for NAMESPACE-URI attribute. Calls {@link #readOnlyError}, as this attribute is r/o
    * for this resource.
    *
    * @param    uri
    *           Not used.
    */
   @Override
   public void setNamespaceURI(character uri)
   {
      readOnlyError("NAMESPACE-URI");
   }


   /**
    * Removes the specified attribute.
    *
    * @param    name
    *           The name of the attribute to remove.
    * @param    uri
    *           The URI of the attribute to remove.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical removeAttribute(character name, character uri)
   {
      if (name.isUnknown())
      {
         invalidArgumentError("REMOVE-ATTRIBUTE");
         return new logical(false);
      }

      if (!isInitialized())
      {
         invalidNodeError("REMOVE-ATTRIBUTE");
         return new logical(false);
      }

      Node node = getNode();
      if (node.getNodeType() != Node.ELEMENT_NODE)
      {
         commonDomError("REMOVE-ATTRIBUTE", "Invalid node type");
         return new logical(false);
      }

      Element elementNode = (Element) node;
      String nameStr = name.toStringMessage();
      if (uri.isUnknown())
      {
         elementNode.removeAttribute(nameStr);
      }
      else
      {
         elementNode.removeAttributeNS(uri.toStringMessage(), nameStr);
      }

      return new logical(true);
   }

   /**
    * Removes the specified attribute. This method is not applicable for Dom.
    *
    * @param    index
    *           The 1-based index of the attribute to remove.
    * @param    uri
    *           The URI of the attribute to remove.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical removeAttribute(NumberType index, character uri)
   {
      return new logical(false); //only for SAX
   }

   /**
    * Setter for NAMESPACE-PREFIX attribute.
    *
    * @param    prefix
    *           The new value of the NAMESPACE-PREFIX attribute.
    */
   @Override
   public void setNamespacePrefix(String prefix)
   {
      setNamespacePrefix(new character(prefix));
   }

   /**
    * Setter for NAMESPACE-PREFIX attribute.
    *
    * @param    prefix
    *           The new value of the NAMESPACE-PREFIX attribute.
    */
   @Override
   public void setNamespacePrefix(character prefix)
   {
      readOnlyError("NAMESPACE-PREFIX");
   }

   /**
    * Returns the value of the the NAMESPACE-PREFIX attribute. Only for ELEMENT or ATTRIBUTE,
    * for other attributes return unknown value.
    *
    * @return    See above.
    */
   @Override
   public character getNamespacePrefix()
   {
      if (!isInitialized() || LegacyResource.X_DOCUMENT.equalsIgnoreCase(type()))
      {
         invalidNodeError("NAMESPACE-PREFIX");
         return new character();
      }

      Node currentNode = getNode();
      final String prefix;
      if (currentNode.getNodeType() == Node.ELEMENT_NODE)
      {
         Element elementNode = (Element) currentNode;
         prefix = elementNode.getPrefix();
      }
      else if (currentNode.getNodeType() == Node.ATTRIBUTE_NODE)
      {
         Attr attributeNode = (Attr) currentNode;
         prefix = attributeNode.getPrefix();
      }
      else
      {
         prefix = null;
      }

      return new character(prefix == null ? "" : prefix);
   }

   /**
    * Get the <code>name</code> attribute of handle.
    *
    * @return   See above.
    */
   @Override
   public character getName()
   {
      if (!isInitialized())
      {
         invalidNodeError("NAME");
         return new character();
      }

      Node node = getNode();

      return new character(node.getNodeName());
   }

   /**
    * Process and display invalid node error. It displays appropriate error message.
    *
    * @param    attribute
    *           The read-only attribute.
    */
   protected void invalidNodeError(String attribute)
   {
      if (attribute == null)
      {
         throw new IllegalArgumentException("The attribute is null !");
      }

      String msg = "X-NODEREF must be associated with a valid X-DOCUMENT in order " +
                   "to use it in method %s";
      String err = String.format(msg, attribute.toUpperCase());
      ErrorManager.recordOrShowError(9102, err, false, false, false);
   }

   /**
    * Process and display unknown node error. It displays appropriate error message.
    *
    * @param    attribute
    *           The read-only attribute.
    * @param    message
    *           Error message to display.
    */
   protected void commonDomError(String attribute, String message)
   {
      if (attribute == null)
      {
         throw new IllegalArgumentException("The attribute is null !");
      }

      String msg = "X-NODEREF or X-DOCUMENT %s got an error: %s";
      String err = String.format(msg, attribute.toUpperCase(), message);
      ErrorManager.recordOrShowError(9082, err, false, false, false);
   }

   /**
    * Process and display unsupported xml sub-type error. It displays appropriate error message.
    *
    * @param    type
    *           The read-only attribute.
    */
   protected void unsupportedDomError(String type)
   {
      String msg = "Unsupported XML SUB-TYPE %s";
      String err = String.format(msg, type);
      ErrorManager.recordOrShowError(9171, err, false, true, false);
   }

   /**
    * Reports if this object is not initialized. Initialized means that object is created but
    * doesn't have any xml node.
    *
    * @return   <code>true</code> if object is initialized.
    */
   boolean isInitialized()
   {
      return getNode() != null;
   }

   /**
    * Setting the Element backend object for this XEntity.
    *
    * @param    xnode
    *           The node value to set.
    */
   void setNode(Node xnode)
   {
      this.xnode = xnode;
   }
}
