/*
** Module   : com.goldencode.p2j.xml
** Abstract : SAX reader handler specific features implementation
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVK 20131030 Created initial version.
*/
package com.goldencode.p2j.xml;

import com.goldencode.p2j.util.*;
import org.xml.sax.*;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.*;
import java.io.IOException;
import java.util.*;

import static com.goldencode.p2j.util.ControlFlowOps.ArgValidationErrors;
import static com.goldencode.p2j.xml.SaxEntityImpl.ERR_10511;
import static com.goldencode.p2j.xml.SaxEntityImpl.ERR_10512;

/**
 * This class provides implementations for all of the possible 4gl-sax callbacks.
 *
 * */

class XMLCallbackWrapper
{
   /** The procedure name to be associated with the given handle object.*/
   private final handle procHandle;

   /** The SaxReader instance to be associated with the given handle object.*/
   private final handle saxReader;

   /** is using memptr in characters callback ?. */
   private final boolean isMemptrInCharacterCallback;

   /** is using memptr in ignorableWhitespace callback ?. */
   private final boolean isMemptrInIgnorableWhitespaceCallback;

   /**
    * Default constructor accepting a handle to procedure and SaxReader instance.
    *
    * @param    procHandle
    *           The procedure name to be associated with the given handle object.
    * @param    saxReader
    *           The SaxReader instance to be associated with the given handle object.
    */
   XMLCallbackWrapper(handle procHandle, handle saxReader)
   {
      this.procHandle = procHandle;
      this.saxReader = saxReader;
      ArgValidationErrors res = validateCallback(CallBack.Characters,
                                                 new memptr(),
                                                 new character());
      isMemptrInCharacterCallback = res == ArgValidationErrors.NO_ERROR;

      res = validateCallback(CallBack.IgnorableWhitespace,
                             new memptr(),
                             new character());
      isMemptrInIgnorableWhitespaceCallback = res == ArgValidationErrors.NO_ERROR;
   }

   /**
    * Receive notification of a notation declaration. If same callback is not exists in 4gl code
    * then do nothing.
    *
    * @param    notationDeclaration
    *           NotationDeclaration xml event which provide all information about notation.
    * @throws   XMLStreamException
    */
   void notationDecl(NotationDeclaration notationDeclaration)
   throws XMLStreamException
   {
      invokeProcedure(CallBack.NotationDecl,
                      new character(notationDeclaration.getName()),
                      new character(notationDeclaration.getPublicId()),
                      new character(notationDeclaration.getSystemId()));
   }

   /**
    * Receive notification of the beginning of the document. If same callback is not exists in 4gl
    * code then do nothing.
    *
    * @throws   XMLStreamException
    */
   void startDocument()
   throws XMLStreamException
   {
      invokeProcedure(CallBack.StartDocument);
   }

   /**
    * Receive notification of the ending of the document. If same callback is not exists in 4gl
    * code then do nothing.
    *
    * @throws   XMLStreamException
    */
   public void endDocument()
   throws XMLStreamException
   {
      invokeProcedure(CallBack.EndDocument);
   }

   /**
    * Receive notification of the start of a Namespace mapping. If same callback is not exists in
    * 4gl code then do nothing.
    *
    * @param    prefix
    *           The Namespace prefix being declared.
    * @param    uri
    *           The Namespace URI mapped to the prefix
    * @throws   XMLStreamException
    */
   public void startPrefixMapping(String prefix, String uri)
   throws XMLStreamException
   {
      invokeProcedure(CallBack.StartPrefixMapping,
                      new character(prefix == null ? "" : prefix),
                      new character(uri));
   }

   /**
    * Receive notification of the end of a Namespace mapping. If same callback is not exists in
    * 4gl code then do nothing.
    *
    * @param    prefix
    *           The Namespace prefix being declared.
    * @throws   XMLStreamException
    */
   public void endPrefixMapping(String prefix)
   throws XMLStreamException
   {
      invokeProcedure(CallBack.EndPrefixMapping, new character(prefix));
   }

   /**
    * Receive notification of the start of an element. If same callback is not exists in 4gl code
    * then do nothing.
    *
    * @param    startElement
    *           Provides access to information about start element.
    *           @see StartElement
    * @throws   XMLStreamException
    */
   public void startElement(StartElement startElement)
   throws XMLStreamException
   {
      QName qName = startElement.getName();

      Iterator attributes = startElement.getAttributes();

      SaxAttributes saxAttributes = new SaxAttributesImpl();
      while(attributes.hasNext()) {
         Attribute attribute = (Attribute)attributes.next();

         QName qNameAttr = attribute.getName();
         saxAttributes.insertAttribute(qNameAttr.getLocalPart(), attribute.getValue(),
                                       qNameAttr.getNamespaceURI());
      }

      invokeProcedure(CallBack.StartElement,
                      new character(qName.getNamespaceURI()),
                      new character(qName.getLocalPart()),
                      new character(qName.getPrefix() + qName.getLocalPart()),
                      new handle(saxAttributes));

   }

   /**
    * Receive notification of the end of an element. If same callback is not exists in 4gl code
    * then do nothing.
    *
    * @param    endElement
    *           Provides access to information about end element.
    *           @see EndElement
    * @throws   XMLStreamException
    */
   public void endElement(EndElement endElement)
   throws XMLStreamException
   {
      QName qName = endElement.getName();
      invokeProcedure(CallBack.EndElement,
                      new character(qName.getNamespaceURI()),
                      new character(qName.getLocalPart()),
                      new character(qName.getPrefix() + qName.getLocalPart()));
   }


   /**
    * Receive notification of character data inside an element. If same callback is not exists in
    * 4gl code then do nothing.
    *
    * @param    characters
    *           Provides access to information about characters element.
    *           @see Characters
    * @throws   XMLStreamException
    */
   public void characters(Characters characters)
   throws XMLStreamException
   {
      String data = characters.getData();
      if (isMemptrInCharacterCallback)
      {
         byte[] arr = data.getBytes();
         invokeProcedure(CallBack.Characters, new memptr(arr), arr.length);
      }
      else
      {
         invokeProcedure(CallBack.Characters, new longchar(data), data.length());
      }
   }

   /**
    * Receive notification of ignorable whitespace in element content. If same callback is not
    * exists in 4gl code then do nothing.
    *
    * @param    characters
    *           Provides access to information about characters element.
    *           @see Characters
    * @throws   XMLStreamException
    */
   public void ignorableWhitespace(Characters characters)
   throws XMLStreamException
   {
      String data = characters.getData();
      if (isMemptrInIgnorableWhitespaceCallback)
      {
         byte[] arr = data.getBytes();
         invokeProcedure(CallBack.IgnorableWhitespace, new memptr(arr), arr.length);
      }
      else
      {
         invokeProcedure(CallBack.IgnorableWhitespace, new longchar(data), data.length());
      }
   }

   /**
    * Receive notification of a processing instruction. If same callback is not
    * exists in 4gl code then do nothing.
    *
    * @param    pi
    *           An interface that describes the data found in processing instructions.
    *           @see ProcessingInstruction
    * @throws   XMLStreamException
    */
   public void processingInstruction(ProcessingInstruction pi)
   throws XMLStreamException
   {
      invokeProcedure(CallBack.ProcessingInstruction, new character(pi.getTarget()),
                      new character(pi.getData()));
   }

   /**
    * Receive notification of a warning messages. If same callback is not
    * exists in 4gl code then do nothing.
    *
    * @param    msg
    *           warning message
    * @throws   XMLStreamException
    */
   public void warning(String msg)
   throws XMLStreamException
   {
      invokeProcedure(CallBack.Warning, new character(msg));
   }

   /**
    * Receive notification of a error messages. If same callback is not
    * exists in 4gl code then do nothing.
    *
    * @param    message
    *           error message
    * @throws   XMLStreamException
    */
   public void error(String message)
   throws XMLStreamException
   {
      invokeProcedure(CallBack.Error, new character(message));
   }

   /**
    * Receive notification of a fatal errors. If same callback is not
    * exists in 4gl code then do nothing.
    *
    * @param    e
    *           Fatal error exception
    * @throws   XMLStreamException
    */
   public void fatalError(Exception e)
   throws XMLStreamException
   {
      invokeProcedure(CallBack.FatalError, new character(e.getMessage()));
   }

   /**
    * Method can invoke procedure callback and valid input parameters. If specified callback is
    * not exists then do nothing.
    *
    * @param    callBack
    *           Callback entity.
    * @param    args
    *           Array of input/output arguments.
    * @throws   XMLStreamException
    */
   private void invokeProcedure(CallBack callBack, Object... args)
   throws XMLStreamException
   {
      if (procHandle == null || !ProcedureManager.isProcedure(procHandle))
      {
         return;
      }

      ArgValidationErrors result = validateCallback(callBack, args);
      if (result == null)
      {
         return;
      }

      if (result == ArgValidationErrors.INCORRECT_COUNT)
      {
         throw new MismatchedParametersException(callBack.name(), ERR_10511);
      }
      if (result != ArgValidationErrors.NO_ERROR)
      {
         throw new MismatchedParametersException(callBack.name(), ERR_10512);
      }

      SelfManager.pushSelf(saxReader);
      try
      {
         ControlFlowOps.invokeInWithMode(callBack.name(), procHandle, callBack.getModes(), args);
      }
      catch (Exception ignored) // just ignore exception
      {
      }
      finally
      {
         SelfManager.popSelf();
      }
   }

   private ArgValidationErrors validateCallback(CallBack callBack, Object... args)
   {
      return ControlFlowOps.reachableInternalEntry(procHandle,
                                                   callBack.name(), false,
                                                   callBack.getModes(),
                                                   args);
   }

   /**
    * List of possible callbacks of SAX-READER object:
    * <ul>
    * <li>{@link #StartDocument} -Start Document callback.</li>
    * <li>{@link #EndDocument} - End Document callback.</li>
    * <li>{@link #ResolveEntity} - ResolveEntity callback.</li>
    * <li>{@link #NotationDecl} - Notation Declaration callback.</li>
    * <li>{@link #UnparsedEntityDecl} - Unparsed Entity Declaration callback.</li>
    * <li>{@link #StartPrefixMapping} - Start Prefix Mapping callback.</li>
    * <li>{@link #EndPrefixMapping} - End Prefix Mapping callback.</li>
    * <li>{@link #StartElement} - Start Element callback.</li>
    * <li>{@link #EndElement} - End Element callback.</li>
    * <li>{@link #Characters} - Characters callback.</li>
    * <li>{@link #IgnorableWhitespace} - Ignorable Whitespace callback.</li>
    * <li>{@link #ProcessingInstruction} - Processing Instruction callback.</li>
    * <li>{@link #Warning} - Warning callback.</li>
    * <li>{@link #Error} - Error callback.</li>
    * <li>{@link #FatalError} - FatalError callback.</li>
    * </ul>
    */
   enum CallBack
   {
      StartDocument(),
      EndDocument(),
      ResolveEntity("IIOO"),  //maybe memptr
      NotationDecl("III"),
      UnparsedEntityDecl("IIII"),
      StartPrefixMapping("II"),
      EndPrefixMapping("I"),
      StartElement("IIII"),
      EndElement("III"),
      Characters("II"),
      IgnorableWhitespace("II"),
      ProcessingInstruction("II"),
      Warning("I"),
      Error("I"),
      FatalError("I");

      /** defined input/output modes for procedure*/
      private final String modes;

      /** default constructor*/
      CallBack()
      {
         this(null);
      }

      /**
       * Special constructor accepting array with possible input type parameters.
       *
       * @param   modes
       *          String of possible input/output parameters type, e.g "I", "O", "U".
       */
      CallBack(String modes)
      {
         this.modes = modes;
      }

      /**
       * Gets the input/output parameter modes.
       *
       * @return   see above.
       */
      String getModes()
      {
         return modes;
      }
   }

   /**
    * Special error throws if some parameter mismatch type.
    */
   static class MismatchedParametersException extends XMLStreamException
   {
      /** name of callback.*/
      final String callbackName;
      /** 4gl error code.*/
      final int errorCode;
      /**
       * Create a new MismatchedParametersException.
       *
       * @param   callbackName
       *          The callback name.
       * @param   errorCode
       *          4gl error code.
       */
      MismatchedParametersException(String callbackName, int errorCode)
      {
         this.callbackName = callbackName;
         this.errorCode = errorCode;
      }
   }
}
